using System.Linq;
using UnityEngine;
using UnityEngine.Events;

/*
 * this plugin is still in beta and dependaing on the feedback i get will be updated to support more features
 * what i would like to add:
 * 1 - a nicer Ui 
 * 2 - a dedicated tab for IK and FK
 * 3 - a better way to move the controllers
*/
namespace Ruvik
{
    class PosingHelper : MVRScript
    {
        protected UIDynamicButton LinkButton;
        protected UIDynamicButton ResetButton;
        protected UIDynamicButton switchMovmentButton;

        static Color pressedColor = new Color(0.3f, 1.0f, 0.3f, 1);
        static Color unPressedColor = new Color(0.839f, 0.839f, 0.839f, 1);

        FreeControllerV3[] knees;
        FreeControllerV3[] hands;
        FreeControllerV3[] feet;
        FreeControllerV3[] thighs;
        FreeControllerV3[] elbows;
        FreeControllerV3[] arms;
        FreeControllerV3 hip;
        FreeControllerV3 chest;
        FreeControllerV3 head;


        static FreeControllerV3[] selected;
        static SelectionButton[] selectionButtons;
        FreeControllerV3.PositionState[] currentPStates;
        FreeControllerV3.RotationState[] currentRStates;

        JSONStorableFloat sensitivity= null;

        //all the other stuff
        void InitControllers()
        {
            knees = containingAtom.freeControllers.Where(fc => fc.name.Contains("KneeControl")).ToArray();
            hands = containingAtom.freeControllers.Where(fc => fc.name.Contains("HandControl")).ToArray();
            feet = containingAtom.freeControllers.Where(fc => fc.name.Contains("FootControl")).ToArray();
            thighs = containingAtom.freeControllers.Where(fc => fc.name.Contains("ThighControl")).ToArray();
            elbows = containingAtom.freeControllers.Where(fc => fc.name.Contains("ElbowControl")).ToArray();
            arms = containingAtom.freeControllers.Where(fc => fc.name.Contains("ArmControl")).ToArray();
            hip = containingAtom.freeControllers[1];
            chest = containingAtom.freeControllers[3];
            head = containingAtom.freeControllers[4];
        }
        void InitSelectionPairs()
        {
            selectionButtons = new SelectionButton[15]
            {
                new SelectionButton(false,"head",head),
                new SelectionButton(false,"right shoulder",arms[0]),
                new SelectionButton(false,"left shoulder", arms[1]),
                new SelectionButton(false,"right elbow", elbows[0]),
                new SelectionButton(false,"left elbow", elbows[1]),
                new SelectionButton(false,"right hand", hands[0]),
                new SelectionButton(false,"left hand",  hands[1]),
                new SelectionButton(false,"hip", hip),
                new SelectionButton(false,"right thigh", thighs[0]),
                new SelectionButton(false,"left thigh", thighs[1]),
                new SelectionButton(false,"right knee", knees[0]),
                new SelectionButton(false,"left knee", knees[1]),
                new SelectionButton(false,"right foot", feet[0]),
                new SelectionButton(false,"left foot", feet[1]),
                new SelectionButton(false,"chest", chest),
            };
        }
        void CreateSelectionPairButtons()
        {
            {
                var btn = CreateSelectionButton(ref selectionButtons[0], false);
                CreateSpacer(true).height = btn.height;
                CreateSpacer(true).height = 5;
                CreateSpacer(false).height = 5;
            }
            {
                var btn = CreateSelectionButton(ref selectionButtons[1], false);
                var btn2 = CreateSelectionButton(ref selectionButtons[2], true);
                CreateSpacer(true).height = 5;
                CreateSpacer(false).height = 5;
            }
            {
                var btn = CreateSelectionButton(ref selectionButtons[3], false);
                var btn2 = CreateSelectionButton(ref selectionButtons[4], true);
                CreateSpacer(true).height = 5;
                CreateSpacer(false).height = 5;
            }
            {
                var btn = CreateSelectionButton(ref selectionButtons[5], false);
                var btn2 = CreateSelectionButton(ref selectionButtons[6], true);
                CreateSpacer(true).height = 5;
                CreateSpacer(false).height = 5;
            }
            {
                var btn = CreateSelectionButton(ref selectionButtons[14], false);
                CreateSpacer(true).height = btn.height;
                CreateSpacer(true).height = 5;
                CreateSpacer(false).height = 5;
            }
            {
                var btn = CreateSelectionButton(ref selectionButtons[7], false);
                CreateSpacer(true).height = btn.height;
                CreateSpacer(true).height = 5;
                CreateSpacer(false).height = 5;
            }
            {
                var btn = CreateSelectionButton(ref selectionButtons[8], false);
                var btn2 = CreateSelectionButton(ref selectionButtons[9], true);
                CreateSpacer(true).height = 5;
                CreateSpacer(false).height = 5;
            }
            {
                var btn = CreateSelectionButton(ref selectionButtons[10], false);
                var btn2 = CreateSelectionButton(ref selectionButtons[11], true);
                CreateSpacer(true).height = 5;
                CreateSpacer(false).height = 5;
            }
            {
                var btn = CreateSelectionButton(ref selectionButtons[12], false);
                var btn2 = CreateSelectionButton(ref selectionButtons[13], true);
                CreateSpacer(true).height = 5;
                CreateSpacer(false).height = 5;
            }

        }
        UIDynamicButton CreateSelectionButton(ref SelectionButton sb, bool right)
        {
            sb.btn = CreateButton(sb.name, right);
            sb.btn.button.onClick.AddListener(sb.action);
            return sb.btn;
        }
        public override void Init()
        {
            sensitivity = new JSONStorableFloat("sensitivity", 1.0f, 0.1f, 3.0f, true, true);
            sensitivity.val = 1.0f;
            pluginLabelJSON.val = "Posing Helper";

            LinkButton = CreateButton("link", false);
            LinkButton.height = 100;
            LinkButton.button.onClick.AddListener(LinkAll());
            ResetButton = CreateButton("reset", true);
            ResetButton.height = 100;
            ResetButton.button.onClick.AddListener(Reset());
            var senSlider = CreateSlider(sensitivity, false);
            CreateSpacer(true).height = senSlider.height;
            CreateSpacer(true).height = 20;
            CreateSpacer(false).height = 20;

            InitControllers();
            InitSelectionPairs();
            CreateSelectionPairButtons();

            CreateSpacer(true).height = 40;
            CreateSpacer(false).height = 40;

            CreateButton("Mirror Right Arm", true).button.onClick.AddListener(MirrorArms(true));
            CreateButton("Mirror Left Arm", false).button.onClick.AddListener(MirrorArms(false));
            CreateButton("Mirror Right Leg", true).button.onClick.AddListener(MirrorLegs(true));
            CreateButton("Mirror left Leg", false).button.onClick.AddListener(MirrorLegs(false));
        }
        protected void Update()
        {
            if(Input.GetKeyDown(KeyCode.Space))
            {
                currentPStates = new FreeControllerV3.PositionState[selected.Length];
                currentRStates = new FreeControllerV3.RotationState[selected.Length];
                for (int i = 0; i < selected.Length; i++)
                {
                    currentPStates[i] = selected[i].currentPositionState;
                    currentRStates[i] = selected[i].currentRotationState;
                }
            }
            else if(Input.GetKey(KeyCode.Space))
            {
                float x, y;
                Vector3 moveby;
                x = -Input.GetAxis("Mouse X") * Time.deltaTime * sensitivity.val;
                y = Input.GetAxis("Mouse Y") * Time.deltaTime * sensitivity.val;
                moveby = Input.GetKey(KeyCode.LeftShift) ? new Vector3(0, 0, x) : new Vector3(x, y, 0);
                MoveController(moveby,!Input.GetKey(KeyCode.LeftControl));
            }
            else if(Input.GetKeyUp(KeyCode.Space))
            {
                for (int i = 0; i < selected.Length; i++)
                {
                    selected[i].currentPositionState = currentPStates[i];
                    selected[i].currentRotationState = currentRStates[i];
                }
            }
        }
        //movement and mirroring operations
        public void MoveController(Vector3 vec3,bool movingPos)
        {
            if (movingPos)
            {
                for (int i = 0; i < selected.Length; i++)
                {
                    selected[i].currentPositionState = FreeControllerV3.PositionState.On;
                    selected[i].MoveControlRelatve(vec3);
                }
            }
            else
            {
                for (int i = 0; i < selected.Length; i++)
                {
                    selected[i].currentRotationState = FreeControllerV3.RotationState.On;
                    selected[i].RotateControl(new Vector3(vec3.y,-vec3.x,vec3.z),1.0f);
                }
            }

        }
        UnityAction MirrorArms(bool right)
        {
            return delegate ()
            {
                FreeControllerV3.PositionState[] pStates = new FreeControllerV3.PositionState[3];
                FreeControllerV3.RotationState[] rStates = new FreeControllerV3.RotationState[3];
                int mirror = right ? 0 : 1; int mirrored = right ? 1 : 0;
                pStates[0] = arms[mirrored].currentPositionState;
                pStates[1] = elbows[mirrored].currentPositionState;
                pStates[2] = hands[mirrored].currentPositionState;
                rStates[0] = arms[mirrored].currentRotationState;
                rStates[1] = elbows[mirrored].currentRotationState;
                rStates[2] = hands[mirrored].currentRotationState;
                arms[mirrored].  currentPositionState = FreeControllerV3.PositionState.On;
                elbows[mirrored].currentPositionState = FreeControllerV3.PositionState.On;
                hands[mirrored]. currentPositionState = FreeControllerV3.PositionState.On;
                arms[mirrored].  currentRotationState = FreeControllerV3.RotationState.On;
                elbows[mirrored].currentRotationState = FreeControllerV3.RotationState.On;
                hands[mirrored]. currentRotationState = FreeControllerV3.RotationState.On;
                arms[mirrored].transform.position = GetInversePosition(arms[mirror].transform.position);
                elbows[mirrored].transform.position = GetInversePosition(elbows[mirror].transform.position);
                hands[mirrored].transform.position = GetInversePosition(hands[mirror].transform.position);
                arms[mirrored].transform.rotation = GetInverseRotation(arms[mirror].transform.rotation);
                elbows[mirrored].transform.rotation = GetInverseRotation(elbows[mirror].transform.rotation);
                hands[mirrored].transform.rotation = GetInverseRotation(hands[mirror].transform.rotation);
                arms[mirrored].currentPositionState = pStates[0];
                elbows[mirrored].currentPositionState = pStates[1];
                hands[mirrored].currentPositionState = pStates[2];
                arms[mirrored].currentRotationState = rStates[0];
                elbows[mirrored].currentRotationState = rStates[1];
                hands[mirrored].currentRotationState = rStates[2];
            };
        }
        UnityAction MirrorLegs(bool right)
        {
            return delegate ()
            {
                FreeControllerV3.PositionState[] pStates = new FreeControllerV3.PositionState[3];
                FreeControllerV3.RotationState[] rStates = new FreeControllerV3.RotationState[3];
                int mirror = right ? 0 : 1; int mirrored = right ? 1 : 0;
                pStates[0] = thighs[mirrored].currentPositionState;
                pStates[1] = knees[mirrored].currentPositionState;
                pStates[2] = feet[mirrored].currentPositionState;
                rStates[0] = thighs[mirrored].currentRotationState;
                rStates[1] = knees[mirrored].currentRotationState;
                rStates[2] = feet[mirrored].currentRotationState;
                thighs[mirrored].currentPositionState = FreeControllerV3.PositionState.On;
                knees[mirrored].currentPositionState = FreeControllerV3.PositionState.On;
                feet[mirrored].currentPositionState = FreeControllerV3.PositionState.On;
                thighs[mirrored].currentRotationState = FreeControllerV3.RotationState.On;
                knees[mirrored]. currentRotationState = FreeControllerV3. RotationState.On;
                feet[mirrored].  currentRotationState = FreeControllerV3.  RotationState.On;
                thighs[mirrored].transform.position = GetInversePosition(thighs[mirror].transform.position);
                knees[mirrored].transform.position = GetInversePosition(knees[mirror].transform.position);
                feet[mirrored].transform.position = GetInversePosition(feet[mirror].transform.position);
                thighs[mirrored].transform.rotation = GetInverseRotation(thighs[mirror].transform.rotation);
                knees[mirrored]. transform.rotation = GetInverseRotation(knees[mirror]. transform.rotation);
                feet[mirrored].  transform.rotation = GetInverseRotation(feet[mirror].  transform.rotation);
                thighs[mirrored].currentPositionState = pStates[0];
                knees[mirrored].currentPositionState = pStates[1];
                feet[mirrored].currentPositionState = pStates[2];
                thighs[mirrored].currentRotationState = rStates[0];
                knees[mirrored]. currentRotationState = rStates[1];
                feet[mirrored].  currentRotationState = rStates[2];
            };
        }
        Vector3 GetInversePosition(Vector3 pos)
        {
            Vector3 mirrored = new Vector3(hip.transform.position.x - pos.x, pos.y, pos.z);
            return mirrored;
        }
        Quaternion GetInverseRotation(Quaternion rot)
        {
            return new Quaternion(rot.x * -1.0f, rot.y, rot.z, rot.w * -1.0f);
        }
        //linking operations
        public UnityAction LinkAll()
        {
            return delegate ()
            {

                for (int i = 0; i < 2; i++)
                {
                    Link(arms[i], chest);
                    Link(elbows[i], arms[i]);
                    Link(hands[i], elbows[i]);
                    Link(thighs[i], hip);
                    Link(knees[i], thighs[i]);
                    Link(feet[i], knees[i]);
                    //feet[i].SetPositionStateFromString("Lock");
                }
                Link(head, chest);
                Link(chest, hip);
                //pelvis.SetPositionStateFromString("On");
                //pelvis.SetRotationStateFromString("On");
            };
        }
        void Link(FreeControllerV3 a, FreeControllerV3 b)
        {
            a.SelectLinkToRigidbody(b.gameObject.GetComponent<Rigidbody>());
            a.currentPositionState = FreeControllerV3.PositionState.ParentLink;
            a.currentRotationState = FreeControllerV3.RotationState.ParentLink;
        }
        public UnityAction Reset()
        {
            return delegate ()
            {
                foreach (FreeControllerV3 fc in containingAtom.freeControllers)
                {
                    fc.SelectLinkToRigidbody(null);
                }
                head.currentPositionState = FreeControllerV3.PositionState.On;
                arms[0].currentPositionState = FreeControllerV3.PositionState.Off;
                arms[1].currentPositionState = FreeControllerV3.PositionState.Off;
                chest.currentPositionState = FreeControllerV3.PositionState.On;
                elbows[0].currentPositionState = FreeControllerV3.PositionState.Off;
                elbows[1].currentPositionState = FreeControllerV3.PositionState.Off;
                hands[0].currentPositionState = FreeControllerV3.PositionState.On;
                hands[1].currentPositionState = FreeControllerV3.PositionState.On;
                hip.currentPositionState = FreeControllerV3.PositionState.On;
                thighs[0].currentPositionState = FreeControllerV3.PositionState.Off;
                thighs[1].currentPositionState = FreeControllerV3.PositionState.Off;
                knees[0].currentPositionState = FreeControllerV3.PositionState.Off;
                knees[1].currentPositionState = FreeControllerV3.PositionState.Off;
                feet[0].  currentPositionState = FreeControllerV3.PositionState.On;
                feet[1].  currentPositionState = FreeControllerV3.PositionState.On;
                head.     currentRotationState = FreeControllerV3.RotationState.On;
                arms[0].  currentRotationState = FreeControllerV3.RotationState.Off;
                arms[1].  currentRotationState = FreeControllerV3.RotationState.Off;
                chest.    currentRotationState = FreeControllerV3.RotationState.On;
                elbows[0].currentRotationState = FreeControllerV3.RotationState.Off;
                elbows[1].currentRotationState = FreeControllerV3.RotationState.Off;
                hands[0]. currentRotationState = FreeControllerV3.RotationState.On;
                hands[1]. currentRotationState = FreeControllerV3.RotationState.On;
                hip.      currentRotationState = FreeControllerV3.RotationState.On;
                thighs[0].currentRotationState = FreeControllerV3.RotationState.Off;
                thighs[1].currentRotationState = FreeControllerV3.RotationState.Off;
                knees[0]. currentRotationState = FreeControllerV3.RotationState.Off;
                knees[1]. currentRotationState = FreeControllerV3.RotationState.Off;
                feet[0].  currentRotationState = FreeControllerV3.RotationState.On;
            };
        }
        // classes and thier functions
        class SelectionButton
        {
            public string name;
            public UnityAction action;
            public bool pressed;
            public FreeControllerV3 selectableFc;
            public UIDynamicButton btn;
            public SelectionButton(bool _pressed, string _name, FreeControllerV3 fc)
            {
                pressed = _pressed;
                name = _name;
                selectableFc = fc;
                action = delegate ()
                {
                    if (Input.GetKey(KeyCode.LeftShift) || Input.GetKey(KeyCode.RightShift))
                    {
                        SetSelectState();
                    }
                    else
                    {
                        foreach (var sb in selectionButtons)
                        {
                            sb.UnSelect();
                        }
                        ClearSelected();
                        Select();
                    }
                };
            }
            void SetSelectState()
            {
                if (pressed == true)
                {
                    UnSelect();
                    RemoveFromSelected(selectableFc);
                }
                else
                    Select();

            }
            void Select()
            {
                pressed = true;
                btn.buttonColor = pressedColor;
                AddToSelected(selectableFc);
            }
            void UnSelect()
            {
                pressed = false;
                btn.buttonColor = unPressedColor;
            }
        }
        static public void RemoveFromSelected(FreeControllerV3 fc)
        {
            int removeAt = 0;
            for (int i = 0; i < selected.Length; i++)
            {
                if (selected[i].name == fc.name)
                    removeAt = i;
            }
            var temp = new FreeControllerV3[selected.Length - 1];
            int oldi = 0, newi = 0;
            while (oldi < selected.Length)
            {
                if (oldi != removeAt)
                {
                    temp[newi] = selected[oldi];
                    newi++;
                }
                oldi++;
            }
            selected = temp;
        }
        static public void AddToSelected(FreeControllerV3 fc)
        {
            var temp = new FreeControllerV3[selected.Length + 1];
            for (int i = 0; i < selected.Length; i++)
            {
                temp[i] = selected[i];
            }
            temp[temp.Length - 1] = fc;
            selected = temp;
        }
        static public void ClearSelected()
        {
            selected = new FreeControllerV3[0];
        }
    }
}
